# queries to do with module/class/function relationships
from __future__ import generators
from getTypeOf import getTypeOf, getTypeOfExpr
from bike.parsing.visitor import walk, walkSourceNodes
from bike.parsing.newstuff import getRegexMatchesInPythonPath, getSourceNode
from bike.query.common import MatchFinder, walkLinesContainingStrings, getScopeForLine
from bike import log
from bike.parsing.fastparserast import Module
import re

# first name is a child of scope
# (e.g. a method in a class, a nested function in a function, a class in a module)
def getChildByName(scope,name):
    matches = [c for c in scope.getChildNodes() if c.name == name]
    if matches != []:
        return matches[0]

def getAllClassesInHierarchy(root,classobj):
    rootClasses = getRootClassesOfHierarchy(classobj)
    print >>log.progress,"Getting root classes"
    relatedClasses = [] + rootClasses
    print >>log.progress,"Getting root classes - done"
    for rootClass in rootClasses:
        print >>log.progress,"Getting all derived classes for ",rootClass
        relatedClasses += getAllDerivedClasses(rootClass)
    return relatedClasses

def getRootClassesOfHierarchy(klass):
    if klass is None:  # i.e. dont have base class in our ast
        return None
    if klass.getBaseClassNames() == []:  # i.e. is a root class
        return[klass]
    else:
        rootclasses = []
        for base in klass.getBaseClassNames():
            baseclass = getTypeOf(klass,base)
            rootclass = getRootClassesOfHierarchy(baseclass)
            if rootclass is None:  # base class not in our ast
                rootclass = [klass]
            rootclasses+=rootclass
        return rootclasses


def getAllDerivedClasses(baseclass):
    state = DerivedClassSearchState()
    return getAllDerivedClasses_impl(baseclass,state)
    
def getAllDerivedClasses_impl(baseclass, state):
    for fname,potentualaliases in state.getCandidateFilesForName(baseclass.name):
        srcnode = getSourceNode(fname)
        walkLinesContainingStrings(srcnode.fastparseroot,
                                   ClassVisitor(baseclass,srcnode,state),
                                   [baseclass.name] + potentualaliases)
    
    return state.subclasses

class DerivedClassSearchState:
    def __init__(self):
        self.classesAndAliases = {}
        self.subclasses = []
        for fname,matchstr in getRegexMatchesInPythonPath("(class\s+[^:( ]+\s*\([^)]+\)|\w+\s+as\s+\w+)"):
            if fname not in self.classesAndAliases:
                self.classesAndAliases[fname] = [matchstr]
            else:
                self.classesAndAliases[fname].append(matchstr)

    # generates (filename,[potentual aliases])
    def getCandidateFilesForName(self,name):
        for filename, matchstrs in self.classesAndAliases.iteritems():
            aliases = []
            gotAMatch = False
            for matchstr in matchstrs:
                match = re.search(name+"\s+as\s+(\w+)",matchstr)
                if match is not None:
                    aliases.append(match.group(1))
                if matchstr.find(name) != -1:
                    gotAMatch = True
            if gotAMatch:
                yield filename,aliases


class ClassVisitor:
    def __init__(self,baseclass,srcnode,state):
        self.baseclass = baseclass
        self.state = state
        self.srcnode = srcnode
        
    def visitClass(self, node):
        classscope = getScopeForLine(self.srcnode, self.lineno)
        
        # create a list of possible aliases for the base class
        # (used to restrict type searches)
        baseclassname = self.baseclass.name
        targetBaseclassNames = [baseclassname] + \
                         scanImportsForAliases(classscope,baseclassname)

        classBaseClasses = classscope.getBaseClassNames()
        for basename in classBaseClasses:
            if basename.split('.')[-1] in targetBaseclassNames and \
                   getTypeOf(classscope,basename) == self.baseclass:
                self.state.subclasses.append(classscope)
                getAllDerivedClasses_impl(classscope,self.state)


# looks for import name as foo, returns [foo,..etc..]
def scanImportsForAliases(scope,name):
    aliases = []        
    while scope:
        for lineno in scope.getImportLines():
            line = scope.module.getSourceNode().getLines()[lineno-1]
            match = re.search(name+"\s+as\s+(\w+)",line)
            if match is not None:
                aliases.append(match.group(1))
        if not isinstance(scope,Module):
            scope = scope.getParent()
        else:
            scope = None
    return aliases
