import unittest
import os
import os.path
from mock import Mock
from bike.parsing.extended_ast import Package, Source, PackageAlreadyExistsException, Root, getRoot, resetRoot
from parsing.utils import fqn_rcar, fqn_rcdr
import re

filesToDelete = None
dirsToDelete = None

class BRMTestCase(unittest.TestCase):
    def setUp(self):
        try: os.makedirs(tmproot)
        except: pass
        os.chdir(tmproot)
        
        resetRoot(Root(None,None,[tmproot]))
        getRoot().unittestmode = True
        global filesToDelete
        global dirsToDelete
        filesToDelete = []
        dirsToDelete = []

    def tearDown(self):
        global filesToDelete
        global dirsToDelete

        for path in filesToDelete:
            try: os.remove(path)
            except: pass
        filesToDelete = []
            
        for path in dirsToDelete:
            try: os.removedirs(path)
            except: pass
        dirsToDelete = []

        os.chdir("..")
        try: os.removedirs(tmproot)
        except: pass



tmproot = os.path.abspath("tmproot")
tmpfile = os.path.join(tmproot, "bicyclerepairman_tmp_testfile.py")
tmpmodule = "bicyclerepairman_tmp_testfile"


def writeFile(filename, src):
    f = open(filename, "w+")
    f.write(src)
    f.close()
    filesToDelete.append(filename)

def readFile(filename):
    f = open(filename)
    src = f.read()
    f.close()
    return src

def writeTmpTestFile(src):
    try:
        os.makedirs(tmproot)
    except OSError:
        pass
    writeFile(tmpfile, src)

def readTmpTestFile():
    return readFile(tmpfile)

def deleteTmpTestFile():
    os.remove(tmpfile)
    os.removedirs(tmproot)


pkgstructureRootDir = tmproot
pkgstructureBasedir = os.path.join(pkgstructureRootDir, "a")
pkgstructureChilddir = os.path.join(pkgstructureBasedir, "b")
pkgstructureFile1 = os.path.join(pkgstructureBasedir, "foo.py")
pkgstructureFile2 = os.path.join(pkgstructureChilddir, "bah.py")

def createPackageStructure(src1, src2):
    try: os.makedirs(pkgstructureChilddir)
    except: pass
    writeFile(os.path.join(pkgstructureBasedir, "__init__.py"), "#")
    writeFile(os.path.join(pkgstructureChilddir, "__init__.py"), "#")
    writeFile(pkgstructureFile1, src1)
    writeFile(pkgstructureFile2, src2)

def removePackageStructure():
    os.remove(os.path.join(pkgstructureBasedir, "__init__.py"))
    os.remove(os.path.join(pkgstructureChilddir, "__init__.py"))
    os.remove(pkgstructureFile1)
    os.remove(pkgstructureFile2)
    os.removedirs(pkgstructureChilddir)


def createAST(src):
    ast = getRoot()
    s = Source.createFromString(tmpfile, tmpmodule, src)
    ast.addSource(s)
    return s

def createAST(src):
    #ast = getRoot()
    #s = Source.createFromString(tmpfile, tmpmodule, src)
    #ast.addSource(s)
    from bike.parsing.newstuff import getSourceNode
    writeFile(tmpfile,src)
    return getSourceNode(tmpfile)


def createSourceNodeAt_old(root, src, fqn):
    return createSourceNodeAt(src,fqn)

#DELETE ME!
def createSourceNodeAt_DELETEME(src, fqn):
    root = getRoot()
    node = root
    modname = fqn_rcar(fqn)
    packagefqn = fqn_rcdr(fqn)
    if packagefqn != "":
        for packagename in packagefqn.split("."):
            try:
                node.addPackage(Package(packagename, packagename))
            except PackageAlreadyExistsException:
                pass
            node = node.getPackage(packagename)

    # mock up the file path
    path = os.path.join(*(packagefqn.split(".")+[modname+".py"]))
    sourcenode = Source.createFromString(path, modname, src)
    node.addSource(sourcenode)
    return root

def createSourceNodeAt(src, fqn):
    modname = fqn_rcar(fqn)
    packagefqn = fqn_rcdr(fqn)
    dirpath = os.path.join(*packagefqn.split("."))
    filepath = os.path.join(dirpath,modname+".py")
    try: os.makedirs(dirpath)
    except: pass
    dirsToDelete.append(dirpath)

    # add the __init__.py files
    path = "."
    for pathelem in packagefqn.split("."):
        path = os.path.join(path,pathelem)
        initfile = os.path.join(path,"__init__.py")
        writeFile(initfile,"#")
        filesToDelete.append(initfile)
    writeFile(filepath,src)
    filesToDelete.append(filepath)
    return getRoot()
    
    

# uses compiler visitor interface along with some dynamic magic
# to get a node out of the ast tree obtained from the src
# N.B. index refers to the number of nodes to count before returning.
# e.g. for second function, nodename=Function, index=1
def getASTNodeFromSrc(src, nodename, index = 0):
    ast = getRoot()
    ast.addSource(Source.createFromString("foo.py", "foo", src))
    return getASTNode(ast, nodename, index)

def getASTNode(ast, nodename, index = 0):
    import compiler
    class FindNode:
        def __init__(self, nodename):
            self.nodename = nodename
            self.index = 0
        def __getattr__(self, name):
            if(name == "visit"+self.nodename):
                return self.visitNode
            else:
                assert name != "node"
                raise AttributeError, name
        def visitNode(self, node, scope = 0):
            if self.index == index:
                self.node = node
            self.index+=1
            for child in node.getChildNodes():
                self.visit(child)

    f = FindNode(nodename)
    compiler.walk(ast, f)
    return f.node



def getMatchingASTNode(ast, nodename, **predicates):
    assert ast
    assert nodename
    import compiler
    class FindNode:
        def __init__(self, nodename):
            self.nodename = nodename
            self.node = None
        def __getattr__(self, name):
            if(name == "visit"+self.nodename):
                return self.visitNode
            else:
                assert name != "node"
                raise AttributeError, name

        def _nodeMatchesPredicate(self, node, name, value):
            return hasattr(node, name)and getattr(node, name) == value

        def visitNode(self, node):
            thisIsTheNode = 1
            for name, value in predicates.items():
                if not self._nodeMatchesPredicate(node, name, value):
                    thisIsTheNode = 0
            if thisIsTheNode:
                self.node = node
            else:
                for child in node.getChildNodes():
                    self.visit(child)

    f = FindNode(nodename)
    compiler.walk(ast, f)
    return f.node

# takes the leading whitespace out of a multi line comment.
# means you can imbed """
#                     text like
#                     this
#                     """
# in your code, and it will come out
#"""text like
#this"""
def trimLines(src):
    lines = src.splitlines(1)[1:]
    tabwidth = re.match("\s*",lines[0]).end(0)
    newlines = []
    for line in lines:
        if line == "\n" or line == "\r\n":
            newlines.append(line)
        else:
            newlines.append(line[tabwidth:])
    return "".join(newlines)

